#!/bin/env groovy

package org.ids_mannheim

import groovy.cli.Option
import groovy.cli.Unparsed
import groovy.cli.picocli.CliBuilder
@GrabConfig(systemClassLoader = true)
@Grab('org.tukaani:xz:1.9')
// should be imported even if not used directly
@Grab('org.codehaus.gpars:gpars:1.2.1')
@Grab('org.apache.commons:commons-compress:1.22')
@Grab('info.picocli:picocli:4.6.3')

import groovyx.gpars.GParsPool

// @GrabConfig(systemProperties = "groovy.grape.enable=false")

import org.apache.commons.compress.compressors.CompressorStreamFactory

import java.util.logging.Logger

class FilterKeys {

    static tag = "FilterKeys"
    static final int maxSpecialKeys = 1000

    static interface pseudonymizeArgs {
        @Option(shortName = 'k', description = "pseudonymization key, use multiple times for multiple keys")
        String[] keys()

        @Option(shortName = 'h', description = "print this help message")
        boolean help()

        @Unparsed(description = "pseudonymized frequency lists to compute essential keys for")
        List files()
    }

    static def compressorOutputStream(fname) {
        def compresserTypes = ["gz": CompressorStreamFactory.GZIP, "xz": CompressorStreamFactory.XZ, "bzip": CompressorStreamFactory.BZIP2]
        String extension = fname.substring(fname.lastIndexOf(".") + 1)
        def type = compresserTypes[extension]
        if (type) return new PrintStream(new BufferedOutputStream(new CompressorStreamFactory().createCompressorOutputStream(type, new BufferedOutputStream(new FileOutputStream(fname))))) else return new PrintStream(new BufferedOutputStream(new FileOutputStream(fname)))
    }

    static def readKey(fname, log) {
        def myArray = new ArrayList<String>()
        log.info("Reading key ${fname} ...")
        def input = new CompressorStreamFactory().createCompressorInputStream(new BufferedInputStream(new FileInputStream(fname)))
        input.splitEachLine("\t") { fields -> myArray[Integer.valueOf(fields[fields.size() - 1]) + maxSpecialKeys] = fields[0] }
        log.info("Done reading key ${fname}.")
        return myArray
    }


    static void main(String[] args) {
        System.setProperty("java.util.logging.SimpleFormatter.format", '[%1$tF %1$tT]:%4$s: %5$s%n')
        Logger log = Logger.getLogger("")

        CliBuilder cli = new CliBuilder(name: tag, width: 100)

        def options = cli.parseFromSpec(pseudonymizeArgs, args)

        if (options.help() || !options.files() || options.files()[0].startsWith('-')) {
            cli.usage()
            System.exit(-1)
        }

        filterKeys(options.files(), options.keys(), log)
    }

    static String replaceDot(a, replaceWith) {
        retun a.substring(0, a.lastIndexOf(",")) + replaceWith + a.substring(a.lastIndexOf(",") + 1)
    }

    static String foldKeyName(fname, keyname, Logger log) {
        def fold = fname.replaceAll(/.*[^0-9]([0-9]+)\..*$/, '$1')
        if (fold == "") {
            log.severe("File ${fname} contains no fold number")
            System.exit(-1)
        }
        return (keyname.replaceFirst(/\./, "." + fold + '.'))
    }

    static void filterKeys(List<String> files, keys, Logger log) {
        log.info("Filtering ${keys.size()} keys")
        def keyArrays = []

        keys.each { fname -> keyArrays << readKey(fname, log) }

        GParsPool.withPool {
            files.eachParallel(fname -> {
                def myKeys = []
                keys.each { myKeys << new ArrayList<Boolean>() }

                def outNames = keys.collect { foldKeyName(fname, it, log) }
                if (outNames.any { new File(it).exists() }) {
                    log.warning("One of ${outNames} already exists - skipping fold")
                } else {
                    log.info("Filtering ${fname} to ${outNames}")
                    def output_streams = outNames.collect { compressorOutputStream(it) }
                    def input = new CompressorStreamFactory().createCompressorInputStream(new BufferedInputStream(new FileInputStream(fname)))

                    input.splitEachLine("\t") { fields ->
                        if (fields.size() >= keyArrays.size()) {
                            for (int i in 0..(fields.size() - 2)) {
                                if (i % 3 < keyArrays.size()) {
                                    myKeys[i % 3][Integer.valueOf(fields[i]) + maxSpecialKeys] = true
                                }
                            }
                        }
                    }
                    log.info("Done filtering ${fname}")
                    GParsPool.withPool {
                        output_streams.eachWithIndexParallel { PrintStream outStream, int i ->
                            log.info("Printing ${outNames[i]}")

                            for (int k : 0..myKeys[i].size()) {
                                if (myKeys[i][k]) outStream.println("${keyArrays[i][k] - maxSpecialKeys}\t${k - maxSpecialKeys}")
                            }
                            outStream.close()
                            log.info("Done printing ${outNames[i]}")
                        }
                    }
                }
            })
        }
    }

}
