package org.ids_mannheim;

import org.anarres.parallelgzip.ParallelGZIPInputStream;
import org.tukaani.xz.XZInputStream;

import java.io.*;
import java.util.ArrayList;
import java.util.Locale;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static java.lang.String.join;
import static java.lang.Thread.sleep;

public class Worker implements Runnable {
    private static final Pattern new_text_pattern = Pattern.compile("^#\\s+text_id\\s*=\\s*(.+)");
    private static final int MAX_RETRIES = 10;
    private final ArrayList<String> fnames;
    private final BlockingQueue<Integer> queue;
    private final ConcurrentHashMap<String, AtomicInteger> map;
    private final Progressbar etaPrinter;
    private final int ngram_size;
    private final int target_fold;
    private final Logger logger;
    private final WorkerNodePool pool;
    private final boolean with_lemma_and_pos;
    private final boolean downcase_tokens;
    private final boolean addPadding;

    private final DeterministicRandomProvider deterministicRandomProvider;

    public Worker(BlockingQueue<Integer> queue, ArrayList<String> fnames, int ngram_size, int target_fold, int folds,
                  ConcurrentHashMap<String, AtomicInteger> map,
                  boolean with_lemma_and_pos, boolean downcase_tokens, WorkerNodePool pool,
                  Progressbar etaPrinter, Logger logger, boolean addPadding) {
        this.queue = queue;
        this.fnames = fnames;
        this.map = map;
        this.ngram_size = ngram_size;
        this.target_fold = target_fold;
        this.with_lemma_and_pos = with_lemma_and_pos;
        this.pool = pool;
        this.etaPrinter = etaPrinter;
        this.logger = logger;
        this.deterministicRandomProvider = new DeterministicRandomProvider(folds);
        this.downcase_tokens = downcase_tokens;
        this.addPadding = addPadding;
    }

    @Override
    public void run() {
        try {
            int index = queue.take();
            int retries = MAX_RETRIES;
            int texts = 0;
            SlidingWindowQueue slidingWindowQueue = (addPadding ? new PaddedSlidingWindowQueue(ngram_size, s -> FoldedEntry.incr(map, s), with_lemma_and_pos) :
                    new SlidingWindowQueue(ngram_size, s -> FoldedEntry.incr(map, s), with_lemma_and_pos));
            while (index >= 0) {
                String fname = fnames.get(index);
                File current_file = new File(fname);
                long file_size = current_file.length();
                int poolIndex = 0;
                BufferedReader in = null;
                logger.info(String.format("Processing %d/%d %s %s", index, fnames.size(), pool.getHost(poolIndex), current_file.getName()));
                if (fname.matches(".*\\.conllu\\.gz$")) {
                    in = new BufferedReader(new InputStreamReader(new ParallelGZIPInputStream(new FileInputStream(fname))));
                } else if (fname.matches(".*\\.conllu\\.xz$")) {
                    in = new BufferedReader(new InputStreamReader(new XZInputStream(new FileInputStream(fname))));
                } else if (fname.matches(".*\\.conllu?$")) {
                    in = new BufferedReader(new InputStreamReader(new FileInputStream(fname)));
                } else if (fname.matches(".*\\.zip$")) {
                    poolIndex = pool.getNextFree();
                    String[] cmd = {
                            "/bin/sh",
                            "-c",
                            pool.getExec(poolIndex) + "/usr/local/kl/bin/korapxml2conllu " + fname
                    };
                    Process p = Runtime.getRuntime().exec(cmd);
                    in = new BufferedReader(new InputStreamReader(p.getInputStream()));
                } else if (fname.matches(".*\\.(freq|[ct]sv)(\\.[xg]z)?$")) {
                    FileInputStream fileInputStream = new FileInputStream(fname);
                    InputStreamReader inputStreamReader;
                    if (fname.matches(".*\\.gz$")) {
                        logger.info("gz input stream detected for "+fname);
                        inputStreamReader = new InputStreamReader(new ParallelGZIPInputStream(fileInputStream));
                    } else if (fname.matches(".*\\.xz$")) {
                        logger.info("xz input stream detected for "+fname);
                        inputStreamReader = new InputStreamReader(new XZInputStream(fileInputStream));
                    } else {
                        logger.info("uncompressed input stream detected for "+fname);
                        inputStreamReader = new InputStreamReader(fileInputStream);
                    }
                    in = new BufferedReader(inputStreamReader);
                    String line;
                    while((line = in.readLine()) != null) {
                        int tabPos = line.lastIndexOf('\t');
                        if (tabPos >= 0) {
                            FoldedEntry.add(map, line.substring(0, tabPos), Integer.parseInt(line.substring(tabPos+1)));
                        } else {
                            logger.severe(new IllegalArgumentException("Cannot interpret tsv line: "+line).toString());
                            System.exit(-1);
                        }
                    }
                    texts++;
                } else {
                    logger.severe(new IllegalArgumentException("Cannot guess file type for "+fname).toString());
                    System.exit(-1);
                }
                String line;
                int fold = -1;
                while ((line = in.readLine()) != null) {
                    if (line.startsWith("#")) {
                        Matcher matcher = new_text_pattern.matcher(line);
                        if (matcher.find()) {
                            if (fold == target_fold) {
                                slidingWindowQueue.textBreak();
                            }
                            fold = deterministicRandomProvider.getFoldFromTextID(matcher.group(1)) + 1;
                            texts++;
                            if (fold == target_fold) {
                                slidingWindowQueue.reset(fold);
                            }
                        }
                    } else if (fold == target_fold) {
                        String[] strings = line.split("\\t");
                        if (strings.length < 4) {
                            continue;
                        }
                        assert strings.length == 10 : "Error: CoNLL-U Format must have 10 columns in line \""+line+"\"";
                        String token = ( downcase_tokens?
                                Utils.unEscapeEntities(strings[1]).toLowerCase(Locale.ROOT) :
                                Utils.unEscapeEntities(strings[1]));
                        if (with_lemma_and_pos) {
                            String lemma, pos;

                            if (token.equals("\"") || token.equals("'")) {
                                lemma = "\"";
                            } else if (token.equals("&")) {
                                lemma = "&amp;";
                            } else if (token.equals("<")) {
                                lemma = "&lt;";
                            } else if (token.equals(">")) {
                                lemma = "&gt;";
                            } else {
                                lemma = strings[2];
                            }
                            if (token.equals("\"") || token.equals("'") || token.equals("<") || token.equals(">")) {
                                pos = "$(";
                            } else if (token.equals("&")) {
                                pos = "KON";
                            } else {
                                pos = strings[3];
                            }
                            //noinspection ConstantCondition
                            slidingWindowQueue.add(join("\t", token, lemma, pos));
                        } else {
                            slidingWindowQueue.add(token);
                        }
                    }
                }
                if (fold == target_fold) {
                    slidingWindowQueue.textBreak();
                }
                pool.markFree(poolIndex);
                if (texts > 0) {
                    logger.info(pool.getHost(poolIndex) + " finished " + fname + " with " + texts + " texts");
                    etaPrinter.update(file_size);
                    retries = MAX_RETRIES;
                    index = queue.take();
                } else {
                    if (--retries > 0) {
                        logger.warning("Retrying " + fname);
                        sleep(1000);
                    } else {
                        logger.severe("Giving up " + fname);
                        index = queue.take();
                    }
                }
            }
        } catch (InterruptedException | IOException e) {
            Thread.currentThread().interrupt();
        }
    }
}
